"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DetachedFieldIndex = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const util_1 = require("../../util");
const codec_1 = require("../../codec");
const detachedFieldIndexCodec_1 = require("./detachedFieldIndexCodec");
/**
 * The tree index records detached field IDs and associates them with a change atom ID.
 */
class DetachedFieldIndex {
    /**
     * @param name - A name for the index, used as a prefix for the generated field keys.
     * @param rootIdAllocator - An ID allocator used to generate unique field keys.
     */
    constructor(name, rootIdAllocator, idCompressor, options) {
        this.name = name;
        this.rootIdAllocator = rootIdAllocator;
        this.idCompressor = idCompressor;
        // TODO: don't store the field key in the index, it can be derived from the root ID
        this.detachedNodeToField = new Map();
        this.options = options ?? { jsonValidator: codec_1.noopValidator };
        this.codec = (0, detachedFieldIndexCodec_1.makeDetachedNodeToFieldCodec)(idCompressor, this.options);
    }
    clone() {
        const clone = new DetachedFieldIndex(this.name, (0, util_1.idAllocatorFromMaxId)(this.rootIdAllocator.getNextId()), this.idCompressor, this.options);
        (0, util_1.populateNestedMap)(this.detachedNodeToField, clone.detachedNodeToField);
        return clone;
    }
    *entries() {
        for (const [major, innerMap] of this.detachedNodeToField) {
            if (major !== undefined) {
                for (const [minor, entry] of innerMap) {
                    yield { id: { major, minor }, root: entry };
                }
            }
            else {
                for (const [minor, entry] of innerMap) {
                    yield { id: { minor }, root: entry };
                }
            }
        }
    }
    updateMajor(current, updated) {
        const innerCurrent = this.detachedNodeToField.get(current);
        if (innerCurrent !== undefined) {
            this.detachedNodeToField.delete(current);
            const innerUpdated = this.detachedNodeToField.get(updated);
            if (innerUpdated === undefined) {
                this.detachedNodeToField.set(updated, innerCurrent);
            }
            else {
                for (const [minor, entry] of innerCurrent) {
                    (0, core_utils_1.assert)(innerUpdated.get(minor) === undefined, 0x7a9 /* Collision during index update */);
                    innerUpdated.set(minor, entry);
                }
            }
        }
    }
    /**
     * Returns a field key for the given ID.
     * This does not save the field key on the index. To do so, call {@link createEntry}.
     */
    toFieldKey(id) {
        return (0, util_1.brand)(`${this.name}-${id}`);
    }
    /**
     * Returns the FieldKey associated with the given id.
     * Returns undefined if no such id is known to the index.
     */
    tryGetEntry(id) {
        return (0, util_1.tryGetFromNestedMap)(this.detachedNodeToField, id.major, id.minor);
    }
    /**
     * Returns the FieldKey associated with the given id.
     * Fails if no such id is known to the index.
     */
    getEntry(id) {
        const key = this.tryGetEntry(id);
        (0, core_utils_1.assert)(key !== undefined, 0x7aa /* Unknown removed node ID */);
        return key;
    }
    deleteEntry(nodeId) {
        const found = (0, util_1.deleteFromNestedMap)(this.detachedNodeToField, nodeId.major, nodeId.minor);
        (0, core_utils_1.assert)(found, 0x7ab /* Unable to delete unknown entry */);
    }
    /**
     * Associates the DetachedNodeId with a field key and creates an entry for it in the index.
     */
    createEntry(nodeId, count = 1) {
        const root = this.rootIdAllocator.allocate(count);
        if (nodeId !== undefined) {
            for (let i = 0; i < count; i++) {
                (0, core_utils_1.assert)((0, util_1.tryGetFromNestedMap)(this.detachedNodeToField, nodeId.major, nodeId.minor + i) === undefined, 0x7ce /* Detached node ID already exists in index */);
                (0, util_1.setInNestedMap)(this.detachedNodeToField, nodeId.major, nodeId.minor + i, root + i);
            }
        }
        return root;
    }
    encode() {
        return this.codec.encode({
            data: this.detachedNodeToField,
            maxId: this.rootIdAllocator.getNextId(),
        });
    }
    /**
     * Loads the tree index from the given string, this overrides any existing data.
     */
    loadData(data) {
        const detachedFieldIndex = this.codec.decode(data);
        this.rootIdAllocator = (0, util_1.idAllocatorFromMaxId)(detachedFieldIndex.maxId);
        this.detachedNodeToField = detachedFieldIndex.data;
    }
}
exports.DetachedFieldIndex = DetachedFieldIndex;
//# sourceMappingURL=detachedFieldIndex.js.map