"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.FieldKinds = exports.fieldKinds = exports.forbidden = exports.nodeKey = exports.sequence = exports.required = exports.valueChangeHandler = exports.valueFieldEditor = exports.optional = exports.noChangeHandler = void 0;
const core_1 = require("../../core");
const util_1 = require("../../util");
const modular_schema_1 = require("../modular-schema");
const sequence_field_1 = require("../sequence-field");
const optional_field_1 = require("../optional-field");
const multiplicity_1 = require("../multiplicity");
/**
 * ChangeHandler that only handles no-op / identity changes.
 */
exports.noChangeHandler = {
    rebaser: (0, modular_schema_1.referenceFreeFieldChangeRebaser)({
        compose: (changes) => 0,
        invert: (changes) => 0,
        rebase: (change, over) => 0,
    }),
    codecsFactory: () => optional_field_1.noChangeCodecFamily,
    editor: { buildChildChange: (index, change) => (0, util_1.fail)("Child changes not supported") },
    intoDelta: (change, deltaFromChild) => ({}),
    relevantRemovedRoots: (change) => [],
    isEmpty: (change) => true,
};
const optionalIdentifier = "Optional";
/**
 * 0 or 1 items.
 */
exports.optional = new modular_schema_1.FieldKindWithEditor(optionalIdentifier, multiplicity_1.Multiplicity.Optional, optional_field_1.optionalChangeHandler, (types, other) => (other.kind.identifier === exports.sequence.identifier ||
    other.kind.identifier === optionalIdentifier) &&
    (0, modular_schema_1.allowsTreeSchemaIdentifierSuperset)(types, other.types), new Set([]));
exports.valueFieldEditor = {
    ...optional_field_1.optionalFieldEditor,
    set: (ids) => optional_field_1.optionalFieldEditor.set(false, ids),
};
exports.valueChangeHandler = {
    ...exports.optional.changeHandler,
    editor: exports.valueFieldEditor,
};
const requiredIdentifier = "Value";
/**
 * Exactly one item.
 */
exports.required = new modular_schema_1.FieldKindWithEditor(requiredIdentifier, multiplicity_1.Multiplicity.Single, exports.valueChangeHandler, (types, other) => (other.kind.identifier === exports.sequence.identifier ||
    other.kind.identifier === requiredIdentifier ||
    other.kind.identifier === exports.optional.identifier ||
    other.kind.identifier === exports.nodeKey.identifier) &&
    (0, modular_schema_1.allowsTreeSchemaIdentifierSuperset)(types, other.types), new Set());
const sequenceIdentifier = "Sequence";
/**
 * 0 or more items.
 */
exports.sequence = new modular_schema_1.FieldKindWithEditor(sequenceIdentifier, multiplicity_1.Multiplicity.Sequence, sequence_field_1.sequenceFieldChangeHandler, (types, other) => other.kind.identifier === sequenceIdentifier &&
    (0, modular_schema_1.allowsTreeSchemaIdentifierSuperset)(types, other.types), 
// TODO: add normalizer/importers for handling ops from other kinds.
new Set([]));
const nodeKeyIdentifier = "NodeKey";
/**
 * Exactly one identifier.
 */
exports.nodeKey = new modular_schema_1.FieldKindWithEditor(nodeKeyIdentifier, multiplicity_1.Multiplicity.Single, exports.noChangeHandler, (types, other) => (other.kind.identifier === exports.sequence.identifier ||
    other.kind.identifier === requiredIdentifier ||
    other.kind.identifier === exports.optional.identifier ||
    other.kind.identifier === nodeKeyIdentifier) &&
    (0, modular_schema_1.allowsTreeSchemaIdentifierSuperset)(types, other.types), new Set());
/**
 * Exactly 0 items.
 *
 * Using Forbidden makes what types are listed for allowed in a field irrelevant
 * since the field will never have values in it.
 *
 * Using Forbidden is equivalent to picking a kind that permits empty (like sequence or optional)
 * and having no allowed types (or only never types).
 * Because of this, its possible to express everything constraint wise without Forbidden,
 * but using Forbidden can be more semantically clear than optional with no allowed types.
 *
 * For view schema, this can be useful if you need to:
 * - run a specific out of schema handler when a field is present,
 * but otherwise are ignoring or tolerating (ex: via extra fields) unmentioned fields.
 * - prevent a specific field from being used as an extra field
 * (perhaps for some past of future compatibility reason)
 * - keep a field in a schema for metadata purposes
 * (ex: for improved error messaging, error handling or documentation)
 * that is not used in this specific version of the schema (ex: to document what it was or will be used for).
 *
 * For stored schema, this can be useful if you need to:
 * - have a field which can have its schema updated to Optional or Sequence of any type.
 * - to exclude a field from extra fields
 * - for the schema system to use as a default for fields which aren't declared
 * (ex: when updating a field that did not exist into one that does)
 *
 * See {@link emptyField} for a constant, reusable field using Forbidden.
 */
exports.forbidden = new modular_schema_1.FieldKindWithEditor(core_1.forbiddenFieldKindIdentifier, multiplicity_1.Multiplicity.Forbidden, exports.noChangeHandler, 
// All multiplicities other than Value support empty.
(types, other) => exports.fieldKinds.get(other.kind.identifier)?.multiplicity !== multiplicity_1.Multiplicity.Single, new Set());
/**
 * Default field kinds by identifier
 */
exports.fieldKinds = new Map([exports.required, exports.optional, exports.sequence, exports.nodeKey, exports.forbidden].map((s) => [s.identifier, s]));
/**
 * Default FieldKinds with their editor types erased.
 * @internal
 */
exports.FieldKinds = { required: exports.required, optional: exports.optional, sequence: exports.sequence, nodeKey: exports.nodeKey, forbidden: exports.forbidden };
//# sourceMappingURL=defaultFieldKinds.js.map